#include "config.h"
#include "macros.h"
#include "network.h"
#include "readconf.h"
#include "ag.h"
#include "share.h"
#include "redirect.h"
#include "filetransfer.h"
#include "ag_fnc.h"
#include "queue.h"


int srv_s=-1,ft_nld=0;
struct sockaddr_in ag_srv_addr;


/* closes the server connection */
void close_server_conn()
{
	if(srv_s != -1)
		close(srv_s);
	srv_s = -1;
	status = AG_STAT_DISCONNECTED;
	shared = 0;

	/* empty the send queue */
	queue_empty();

	/* trigger event */
	event_action("disconnect", 0);
}


unsigned char *get_ft_bar(int pos, int size)
{
	static unsigned char bar[11];
	int b;

	b = rdiv(pos*11, size)-1;

	memcpy(bar, "          ", 11);

	if(b >= 0 && b < 10)
		bar[b] = '>';
	while(b > 0) bar[--b] = '-';

	return bar;
}


/* fetches a server from the redirect server */
int ag_poll_redirect()
{
	static int s = -1;
	int n;


	if(s == -1)
	{
		context;
		/* fetch a server from the redirect server */
		status = AG_STAT_DISCONNECTED;
		NOTICE("Fetching server from " AG_REDIRECT_SERVER "...");
		/* resolve will block - usually not a problem */
		s = tcp_connect(make_sockaddr(AG_REDIRECT_SERVER, 21), 1);
		if(s == -1)
		{
			NOTICE("Couldn't connect to redirect server: %s", ERRMSG());
			return 0;
		}
	}

	context;

	n = poll_sock(s, 0);
	if(n == -1)
	{
		NOTICE("poll_sock() error: %s", ERRMSG());
		close(s);
		s = -1;
		return 0;
	}

	if(n == 0)
		return 1;

	context;

	if(ag_fetch_redirect(s, &ag_srv_addr))
		status = AG_STAT_CONNECTING;

	close(s);
	s = -1;

	return 1;
}




/*
	receives server packets and sends them off to ag_handle_packet
	(unless the packet is a filetransfer packet)
*/

int ag_poll_server()
{
	static struct sharestruct *share = NULL;
	static unsigned char rbuf[16384];
	static int rbufs=0;
	static struct timeval tv_ping,tv_conn,tv_share,tv;
	int i,n=0,psize;


	if(status != AG_STAT_DISCONNECTED && status != AG_STAT_CONNECTING)
	{
		/* ping the server */
		context;
		gettimeofday(&tv, NULL);
		if(tv_ping.tv_sec+PINGFREQ < tv.tv_sec)
		{
			if(ag_srv_msg(AG_MSG_PING, NULL, 0, QPRIO_HIGH) == -1)
			{
				NOTICE("Error when pinging server: %s", ERRMSG());
				close_server_conn();
				return 0;
			}
			gettimeofday(&tv_ping, NULL);

			NOTICE("Ping! - Status: %d filetransfer%s, %d shared file%s.",
					ft_count, ft_count==1?"":"s", shared, shared==1?"":"s");
			for(i=0;i<MAX_FT;i++)
			{
				if(ft[i].status == FT_ACTIVE)
				{
					int pos = ft[i].fd ? ftell(ft[i].fd) : 0;
					int size = ft[i].filesize;

					/* could happen i guess.. */
					if(pos > size)
						size = pos;

					NOTICE("%s: [%s] %.1f%% (%.2fKB of %.2fKB @ %.2fKB)",
						ft[i].direction == AG_FT_RECV ? "Download" : "  Upload",	/* Download / Upload */
						get_ft_bar(pos, size),										/* Progress bar */
						(float)pos / size * 100.0f,									/* Percent completed */
						(float)pos / 1024.0f,										/* Current position */
						(float)ft[i].filesize / 1024.0f,							/* Total filesize */
						(float)ft[i].speed / 1024.0f);								/* Speed */

					NOTICE("        > %s", ft[i].file);	/* filename */
				}
				else if(ft[i].status != FT_UNUSED)
				{
					NOTICE("%s: Setting up...",
						ft[i].direction == AG_FT_RECV ? "Download" : "  Upload");
				}
			}
			NOTICE("----");
		}
	}

	if(status == AG_STAT_CONNECTING && srv_s == -1)
	{
		/* connect to the server */
		context;
		NOTICE("Connecting to %s:%d...", inet_ntoa(ag_srv_addr.sin_addr), htons(ag_srv_addr.sin_port));

		gettimeofday(&tv_conn, NULL);
		srv_s = tcp_connect(&ag_srv_addr, 1);
		if(srv_s == -1)
		{
			NOTICE("Couldn't connect to server: %s", ERRMSG());
			close_server_conn();
			return 0;
		}
	}


	if(status == AG_STAT_CONNECTING)
	{
		/* request magic */
		context;
		if(ag_srv_msg(AG_MSG_MAGICREQUEST, NULL, 0, QPRIO_HIGH) == -1)
		{
			if(errno == EAGAIN || errno == ENOTCONN)
			{
				gettimeofday(&tv, NULL);
				if(tv_conn.tv_sec+CONNTIMEOUT < tv.tv_sec)
				{
					NOTICE("ZzZzZzzz, trying another server.");
					close_server_conn();
				}
				return 1;
			}
			else
			{
				close_server_conn();
				return 1;
			}
		}
		else
		{
			NOTICE("Connected. Logging in...");
			status = AG_STAT_CONNECTED;
			gettimeofday(&tv_ping, NULL);
			rbufs = 0;

			/* trigger event */
			event_action("connect", 2,
					inet_ntoa(ag_srv_addr.sin_addr),
					int2str(htons(ag_srv_addr.sin_port))
				);
		}
		return 1;
	}


	if(status == AG_STAT_SEND_SHARES)
	{
		struct sharestruct *next_share;
		int cost;

		/* send shares */
		if(share == NULL)
		{
			gettimeofday(&tv_share, NULL);
			share = sharelist;
		}

		/* send shares once every 1-2 seconds */
		gettimeofday(&tv, NULL);
		if(tv_ping.tv_sec+1 > tv.tv_sec)
		{
			cost = 9999;
		}
		else
		{
			gettimeofday(&tv_share, NULL);
			cost = 0;
		}

		context;
		/* if there are any active filetransfers, lower the max cost */
		while(cost < (ft_count ? 200 : 1000))
		{
			if(share) {
				next_share = share->next;

				switch(share->status)
				{
				case SHARE_STAT_NEW:
					ag_send_new_share(share, &cost);
					break;
				case SHARE_STAT_OLD:
					ag_send_old_share(share, &cost);
					shared++;
					break;
				case SHARE_STAT_UPD:
					ag_send_upd_share(share, &cost);
					break;
				case SHARE_STAT_DEL:
					/* This doesn't work, but it doesn't really matter. */
					/*ag_send_old_share(share, *cost);
					ag_send_del_share(share, *cost);*/
					share_del(share);
					save_share_db = 1;
					break;
				}
				share = next_share;
			}

			if(!share)
			{
				status = AG_STAT_LOGGED_IN;
				NOTICE("Shares sent to server.");
				break;
			}
		}
	}


	context;
	/* poll server socket */
	n = poll_sock(srv_s, 0);
	if(n == -1)
	{
		NOTICE("select() error, shutting down server connection.");
		close_server_conn();
		return 0;
	}

	if(n == 0)
		return 1;


	context;
	/* read from server socket */
	n = recv(srv_s, &rbuf[rbufs], sizeof(rbuf)-rbufs, 0);
	if(n <= 0)
	{
		NOTICE("Error while receiving data from server: %s", ERRMSG());
		close_server_conn();
		return 0;
	}
	rbufs += n;

	/* parse packet buffer */
	while(rbufs > 4)
	{
		context;
		/* does the packet have a valid header? */
		if(!AG_VALID_HDR(rbuf))
		{
			DEBUG("Invalid packet header '%c%c'", rbuf[0], rbuf[1]);
			close_server_conn();
			return 0;
		}

		/* do we have a whole packet? */
		if(AG_WORD(&rbuf[2]) <= rbufs-2)
		{
			context;
			psize = AG_WORD(&rbuf[2]);	/* packet size */

			if(rbuf[4] != AG_MSG_FILETRANSFER)
			{
				context;
				/* handles all packets except for filetransfer packets */
				ag_handle_packet(psize-3, rbuf[4], &rbuf[5]);
			}
			else
			{
				context;
				/* handle filetransfer packets */
				i = ft_alloc();
				if(i == -1)
				{
					NOTICE("No free filetransfer slots, aborting transfer!");
					/* hack hack hack */
					if(AG_WORD(&rbuf[5]) < psize-3)
						ft_send_state(AG_FT_STATE_ERROR, &rbuf[7], AG_WORD(&rbuf[5]));
				}
				else
				{
					/* fill the allocated ft_struct with crap */
					if(ft_fillstruct(i, psize, &rbuf[5]) == -1)
					{
						ft_free(i);
						goto fterr;
					}

					context;
					if(ft[i].connecting == AG_FT_CONNECT)
					{
						/* connecting */
						NOTICE("Connecting to %s:%d...", ft[i].ip, ft[i].port);
						ft[i].s = tcp_connect(make_sockaddr(ft[i].ip, ft[i].port), 1);
						if(ft[i].s == -1)
						{
							/* socket error - notify the server and free ft_struct */
							ft_send_state(AG_FT_STATE_ERROR, ft[i].fileid, ft[i].fileid_len);
							ft_free(i);
						}
						else
						{
							ft_new_fd(ft[i].s);
							ft[i].fdset = AG_FD_WRITE;
							ft[i].status = FT_CONNECTING;
						}
					}
					else
					{
						/* listening */
						ft[i].s = tcp_listen(make_sockaddr("0.0.0.0", ft[i].port), 4);
						if(ft[i].s == -1)
						{
							/* listen error - notify server */
							ft_send_state(AG_FT_STATE_ERRBIND, ft[i].fileid, ft[i].fileid_len);
							ft_free(i);
						}
						else
						{
							ft_new_fd(ft[i].s);
							NOTICE("Listening on port %d...", ft[i].port);
							ft[i].fdset = AG_FD_READ;
							ft[i].status = FT_LISTENING;
						}
					}	/* if(ft[i].connecting) */

					/* no errors */
					if(ft[i].status != FT_UNUSED)
					{
						DEBUG("AG_FT_STATE_SETUP");
						ft_count++;
						ft_send_state(AG_FT_STATE_SETUP, ft[i].fileid, ft[i].fileid_len);
						gettimeofday(&ft[i].timeout, NULL);
						if(ft[i].fdset == AG_FD_READ)
							FD_SET(ft[i].s, &ft_read_fd);
						else
							FD_SET(ft[i].s, &ft_write_fd);
					}
fterr:;
				}

			}	/* if(rbuf[4] != AG_MSG_FILETRANSFER) */

			/* re-arrange buffer */
			memcpy(rbuf, rbuf+psize+2, rbufs-(psize+2));
			rbufs -= psize+2;
		}
		else
		{
			/* we don't have a whole packet, break the loop */
			break;
		}
	}	/* if(rbufs > 4) */

	return 1;
}



int ag_poll_filetransfers()
{
	struct sockaddr_in accept_addr;
	int i,r,n,plen,tmp_s,nfd;
	struct timeval tv;

	context;

	nfd = ft_max_fd();

	tv.tv_sec = 1;
	tv.tv_usec = 0;
	n = select(nfd+1, &ft_read_fd, &ft_write_fd, NULL, &tv);
	if(n == -1)
	{
		NOTICE("Fatal select() error: %s", ERRMSG());
		NOTICE("Shutting down all filetransfers!");
		ft_close_all();
		close_server_conn();
		return 0;
	}

	gettimeofday(&tv, NULL);
	for(i=0;i<MAX_FT;i++)
	{
		if(ft[i].status == FT_UNUSED)
			continue;

		/* look for stalled transfers / connect timeouts */
		if(ft[i].timeout.tv_sec+TRANSTIMEOUT < tv.tv_sec)
		{
			NOTICE("Aborting transfer of '%s': %s", ft[i].file, 
					ft[i].status==FT_CONNECTING?"Connection timed out":"Transfer stalled");
			ft_send_state(AG_FT_STATE_NOCONN, ft[i].fileid, ft[i].fileid_len);
			ft_free(i);
			break;
		}

		/* add all sockets back to the fdset if select() returned 0 sets */
		if(n == 0)
		{
			if(ft[i].fdset == AG_FD_READ)
				FD_SET(ft[i].s, &ft_read_fd);
			else
				FD_SET(ft[i].s, &ft_write_fd);
		}
	}

	/* exit if select() returned 0 sets */
	if(n == 0)
		return 1;


	for(i=0;i<MAX_FT;i++)
	{
		if(ft[i].status == FT_UNUSED)
			continue;

		if(ft[i].fdset == AG_FD_READ && !FD_ISSET(ft[i].s, &ft_read_fd))
		{
			/* add the socket back to the read fdset */
			FD_SET(ft[i].s, &ft_read_fd);
			continue;
		}
		if(ft[i].fdset == AG_FD_WRITE && !FD_ISSET(ft[i].s, &ft_write_fd))
		{
			/* add the socket back to the write fdset */
			FD_SET(ft[i].s, &ft_write_fd);
			continue;
		}

		/* reset timeout */
		gettimeofday(&ft[i].timeout, NULL);

		switch(ft[i].status)
		{
		case FT_LISTENING:
			/* someone connected to us */
			context;
			DEBUG("FT_LISTENING");

			/* accept connection */
			n = sizeof(struct sockaddr_in);
			tmp_s = accept(ft[i].s, (struct sockaddr *)&accept_addr, &n);
			if(tmp_s == -1)
			{
				ft_close_conn(i);
				break;
			}

			if(!ip_match(ft[i].ip, &accept_addr.sin_addr))
			{
				NOTICE("Warning: connection from IP '%s' does not match IP mask '%s', dropping...", inet_ntoa(accept_addr.sin_addr), ft[i].ip);
				close(tmp_s);
				continue;
			}

			FD_CLR(ft[i].s, &ft_read_fd);
			ft_del_fd(ft[i].s);
			close(ft[i].s);
			ft[i].s = tmp_s;
			ft_new_fd(ft[i].s);

			if(ft[i].direction == AG_FT_RECV)
			{
				ft[i].fd = ft_fopen(ft[i].file, 1);
				if(!ft[i].fd)
				{
					ft_close_conn(i);
					break;
				}

				ft[i].resume_pos = ftell(ft[i].fd);
				if(ft_send_request(i, ft[i].resume_pos) == -1)
				{
					ft_close_conn(i);
					break;
				}
			}
			ft[i].fdset = AG_FD_WRITE;
			FD_SET(ft[i].s, &ft_write_fd);

			ft[i].status = FT_WAIT;
			break;


		case FT_CONNECTING:
			/* connection established */
			context;
			DEBUG("FT_CONNECTING");

			if(ft[i].direction == AG_FT_RECV)
			{
				ft[i].fd = ft_fopen(ft[i].file, 1);
				if(!ft[i].fd)
				{
					ft_close_conn(i);
					break;
				}

				ft[i].resume_pos = ftell(ft[i].fd);
				if(ft_send_request(i, ft[i].resume_pos) == -1)
				{
					ft_close_conn(i);
					break;
				}
			}
			ft[i].status = FT_WAIT;
			ft[i].fdset = AG_FD_READ;
			FD_CLR(ft[i].s, &ft_write_fd);
			FD_SET(ft[i].s, &ft_read_fd);
			break;


		case FT_WAIT:
			/* waiting for ack / request */
			context;
			DEBUG("FT_WAIT");

			if(!ft[i].buf)
			{
				/* allocate some memory for a receive buffer */
				ft[i].buf = ag_malloc(TRANSBUF);

				if(!ft[i].buf)
				{
					ft_close_conn(i);
					break;
				}
			}

			n = recv(ft[i].s, ft[i].buf, TRANSBUF, 0);
			if(n == -1)
			{
				NOTICE("Aborting transfer of '%s': %s", ft[i].file, ERRMSG());
				if(errno == ECONNRESET)
					ft_send_state(AG_FT_STATE_CLOSED, ft[i].fileid, ft[i].fileid_len);
				else
					ft_send_state(AG_FT_STATE_NOCONN, ft[i].fileid, ft[i].fileid_len);
				ft_free(i);
				break;
			}

			/* handshake */
			if(ft[i].direction == AG_FT_RECV)
			{
				/* got ACK */
				context;
				plen = ft_recv_response(i, n, ft[i].buf);
				if(plen == -1)
				{
					ft_close_conn(i);
					break;
				}

				if(plen > 0 && (n-plen) > 0)
				{
					/* save the overflow (we got more data than the ACK packet...) */
					if(fwrite(ft[i].buf, n-plen, 1, ft[i].fd) == 0)
					{
						NOTICE("Error writing to disk: %s", ERRMSG());
						ft_free(i);
					}
				}
			}
			else
			{
				/* got REQEUST */
				context;

				ft[i].fd = ft_fopen(ft[i].ufile, 0);
				if(!ft[i].fd)
				{
					NOTICE("Error opening '%s': %s", ft[i].ufile, ERRMSG());
					ft_close_conn(i);
					break;
				}

				plen = ft_recv_request(i, n, ft[i].buf);
				if(plen == -1)
				{
					ft_close_conn(i);
					break;
				}

				if(plen > 0)
				{
					if(ft_send_response(i) == -1)
					{
						ft_close_conn(i);
						break;
					}

					/* since we'll be sending data, we'll move this socket to the ft_write_fd ... */
					ft[i].fdset = AG_FD_WRITE;
					FD_CLR(ft[i].s, &ft_read_fd);
					FD_SET(ft[i].s, &ft_write_fd);
				}
			}
			ft[i].status = FT_ACTIVE;

			/* set timers */
			gettimeofday(&ft[i].start, NULL);
			gettimeofday(&ft[i].lastspd, NULL);


			NOTICE("Started %s of '%s'...", ft[i].direction == AG_FT_RECV ? "download" : "upload", ft[i].file);

			if(ft[i].direction == AG_FT_RECV)
				event_action("download_start", 3, conf.download_dir, ft[i].file, ft[i].ip);
			else
				event_action("upload_start", 3, conf.download_dir, ft[i].file, ft[i].ip);

			break;


		case FT_ACTIVE:
			context;

			if(ft[i].direction == AG_FT_RECV)
			{
				context;
				n = recv(ft[i].s, ft[i].buf, TRANSBUF, 0);
				if(n <= 0)
				{
					if(ftell(ft[i].fd) < ft[i].filesize)
					{
						NOTICE("Download of '%s' aborted: %s", ft[i].file, ERRMSG());
						ft_send_state(AG_FT_STATE_CLOSED, ft[i].fileid, ft[i].fileid_len);
					}
					else
					{
						ft_rename(i);
						NOTICE("Download of '%s' completed. Average speed was %.2fKB/sec.", ft[i].file, ft_avg_speed(i));
						ft_send_state(AG_FT_STATE_COMPLETE, ft[i].fileid, ft[i].fileid_len);
						event_action("download_end", 3, conf.download_dir, ft[i].file, float2str(ft_avg_speed(i)));
					}
					ft_free(i);
				}
				else
				{
					if(fwrite(ft[i].buf, n, 1, ft[i].fd) == -1)
					{
						NOTICE("Error while writing '%s': %s", ERRMSG());
						ft_close_conn(i);
					}
					else
					{
						ft_update_speed(i);
					}
				}
			}
			else	/* uploads */
			{
				context;
				r = fread(ft[i].buf, 1, TRANSBUF, ft[i].fd);
				if(r > 0)
				{
					n = send(ft[i].s, ft[i].buf, r, 0);
					if(n == -1)
					{
						if(errno == EAGAIN || errno == EWOULDBLOCK || errno == ENOBUFS)
						{
							/* can't send more - but select() said we could? */
							fseek(ft[i].fd, -r, SEEK_CUR);
							ft_update_speed(i);
						}
						else
						{
							/* send() error */
							NOTICE("Upload of '%s' aborted: %s", ft[i].file, ERRMSG());
							ft_close_conn(i);
						}
					}
					else
					{
						/* if all of the data was not sent, roll back */
						if(n > 0 && n < r)
							fseek(ft[i].fd, -(r-n), SEEK_CUR);
						ft_update_speed(i);
					}
				}
				else if(r == -1)
				{
					NOTICE("Error while reading '%s': %s", ERRMSG());
					ft_close_conn(i);
				}
				else
				{
					NOTICE("Upload of '%s' completed. Average speed was %.2fKB/sec.", ft[i].file, ft_avg_speed(i));
					ft_send_state(AG_FT_STATE_COMPLETE, ft[i].fileid, ft[i].fileid_len);
					event_action("upload_end", 3, conf.download_dir, ft[i].file, float2str(ft_avg_speed(i)));
					ft_free(i);
				}
			}	/* if(ft[i].direction == AG_FT_RECV) - else */
			break;
		}
		context;
	}

	return 1;
}





void ag_main_loop()
{
	context;

	/* init filetransfer structs */
	ft_init();

	while(running)
	{
		if(status == AG_STAT_DISCONNECTED)
		{
			context;
			ag_poll_redirect();
		}

		if(!running)
			break;

		if(status != AG_STAT_DISCONNECTED)
		{
			ag_poll_server();
			share_interval_save_db();
			if(!queue_poll())
				close_server_conn();
		}

		if(!running)
			break;

		context;
		ag_poll_filetransfers();
	}


	if(status && srv_s != -1)
	{
		NOTICE("Logging out...");
		ag_srv_msg(AG_MSG_LOGOUT, NULL, 0, QPRIO_HIGH);
		close_server_conn();
	}

	ft_close_all();
}

