#include "config.h"
#include "macros.h"
#include "share.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netinet/in.h>

/*
	"Ripped" this code from Salvatore Sanfilippo's mp3ai
	http://www.hping.org/mp3ai/

	An mp3 id3tag reader, all in just 1 simple C file.
	Makes it sooooo easy *droool*

	Changes:
	o Added context;'s
	o Removed stuff I didn't need ;)
*/

/* mp3ai.c
 * Copyright(C) 2001 Salvatore Sanfilippo
 * all rights reserved
 *
 * mp3ai is free software, under the terms of the GPL license version 2 */


#define MPGHDR_LEN	4
#define MPGVER_25	0
#define MPGVER_RES	1
#define MPGVER_2	2
#define MPGVER_1	3
#define MPGLAYER_RES	0
#define MPGLAYER_3	1
#define MPGLAYER_2	2
#define MPGLAYER_1	3
#define MPGMODE_STEREO	0
#define MPGMODE_JOINT	1
#define MPGMODE_DUAL	2
#define MPGMODE_SINGLE	3
#define MPGSAMPLES_FOR_FRAME_L1 384		/* Layer I */
#define MPGSAMPLES_FOR_FRAME_L23 1152	/* Layer II and III */
#define MPGSLOT_LEN_L1	4				/* Layer I */
#define MPGSLOT_LEN_L23	1				/* Layer II and III */

#define ID3_LEN		128
#define ID3_TITLESZ	30
#define ID3_ARTISTSZ	30
#define ID3_ALBUMSZ	30
#define ID3_YEARSZ	4
#define ID3_COMMENTSZ	30

struct mp3_info {
	time_t playtime;
	unsigned char ver;
	unsigned int filelen;
	unsigned char layer;
	unsigned int bitrate;
	unsigned char title[ID3_TITLESZ+1];
	unsigned char artist[ID3_ARTISTSZ+1];
	unsigned char album[ID3_ALBUMSZ+1];
	unsigned char track;
};

static int mp3ai_getinfo(unsigned char *filename, struct mp3_info *i, unsigned int size);


/* ------------------- */

int is_fucked(unsigned char *str)
{
	while(*str)
	{
		if(*str < 32)
			return 1;
		str++;
	}
	return 0;
}

void trim_str(unsigned char *str)
{
	int i=strlen(str)-1;

	if(i < 0)
		return;

	while(i >= 0)
	{
		if(str[i] != ' ')
			return;
		str[i] = 0;
	}
}


int mp3_fill_info(struct sharestruct *share, unsigned char *file, unsigned int size)
{
	struct mp3_info i;
	int r=0;

	context;

	if(mp3ai_getinfo(file, &i, size))		/* :) */
		return 0;

	context;

	share->bitrate = i.bitrate;
	share->seclength = (int)i.playtime;

	/*
		mp3ai_getinfo() sometimes returns bogus info
		so we'll look for chars < 32 etc and filter them out

		fortunately this happens very rarely (3 out of 1500 mp3s for me)
	*/
	if(is_fucked(i.title))
		i.title[0] = 0;
	if(is_fucked(i.artist))
		i.artist[0] = 0;
	if(is_fucked(i.album))
		i.album[0] = 0;

	trim_str(i.title);
	trim_str(i.artist);
	trim_str(i.album);

	if(i.title[0]) {
		share->id3v2_title = ag_strdup(i.title); r++;
	}
	if(i.artist[0]) {
		share->id3v2_artist = ag_strdup(i.artist); r++;
	}
	if(i.album[0]) {
		share->id3v2_album = ag_strdup(i.album); r++;
	}
	if(i.track) {
		share->id3v2_track = i.track; r++;
	}

	if(r)
		share->have_id3v2 = 1;

	return 1;
}

/* ------------------- */


static int32 extract_bitfield(unsigned char *h, int start, int end)
{
	int32 hdr;

	context;

	memcpy(&hdr, h, 4);
	hdr = ntohl(hdr);
	hdr = hdr << start;
	hdr = hdr >> (start + (31 - end));
	return hdr;
}

/* MPEG header decoding */
static void mp3ai_dechdr(unsigned char *h, struct mp3_info *i)
{
	static int bitrate_tbl[80] = {
		/* V1 L1 */
		0, 32, 64, 96, 128, 160, 192, 224,
		256, 288, 320, 352, 384, 416, 448, 0,
		/* V1 L2 */
		0, 32, 48, 56, 64, 80, 96, 112,
		128, 160, 192, 224, 256, 320, 384, 0,
		/* V1 L3 */
		0, 32, 40, 48, 56, 64, 80, 96,
		112, 128, 160, 192, 224, 256, 320, 0,
		/* V2[.5] L1 */
		0, 32, 48, 56, 64, 80, 96, 112,
		128, 144, 160, 176, 192, 224, 256, 0,
		/* V2[.5] L2 & L3 */
		0, 8, 16, 24, 32, 40, 48, 56,
		64, 80, 96, 112, 128, 144, 160, 0
	};
	int32 aux;

	context;

	i->ver = extract_bitfield(h, 11, 12);  	/* mpeg version */
	i->layer = extract_bitfield(h, 13, 14); /* mpeg layer */
	aux = extract_bitfield(h, 16, 19); /* bitrate */
	i->bitrate = 0; /* default */
	switch(i->ver) {
	case MPGVER_1:
		switch(i->layer) {
		case MPGLAYER_1:
			i->bitrate = bitrate_tbl[aux];
			break;
		case MPGLAYER_2:
			i->bitrate = bitrate_tbl[aux+16];
			break;
		case MPGLAYER_3:
			i->bitrate = bitrate_tbl[aux+32];
			break;
		}
		break;
	case MPGVER_2:
	case MPGVER_25:
		switch(i->layer) {
		case MPGLAYER_1:
			i->bitrate = bitrate_tbl[aux+48];
			break;
		case MPGLAYER_2:
		case MPGLAYER_3:
			i->bitrate = bitrate_tbl[aux+64];
			break;
		}
		break;
	}
	//i->bitrate *= 1000; /* the bitrate table is /1000 */

	context;

	/* Extimate the playtime and number of frames */
	if(i->bitrate)
		i->playtime = (i->filelen/1000)/(i->bitrate/8);
}

/* ID3 V1.1 tag decoding */
static void mp3ai_dectag(unsigned char *t, struct mp3_info *i)
{
	context;

	memcpy(i->title, t+3, ID3_TITLESZ);
	i->title[ID3_TITLESZ] = '\0';
	memcpy(i->artist, t+33, ID3_ARTISTSZ);
	i->artist[ID3_ARTISTSZ] = '\0';
	memcpy(i->album, t+63, ID3_ALBUMSZ);
	i->album[ID3_ALBUMSZ] = '\0';

	/* track number */
	if (t[125] == 0 && t[126] != 0)
		i->track = t[126];
	else
		i->track = 0;
	return;
}

/* Try to figure if the 4 bytes data pointed by h is an mpeg header */
static int seems_mpeg_hdr(unsigned char *h)
{
	/* invalid frame sync */
	if (extract_bitfield(h, 0, 10) != 0x7FF)
		return 0;
	/* invalid version */
	if (extract_bitfield(h, 11, 12) == 0x01)
		return 0;
	/* invalid layer */
	if (extract_bitfield(h, 13, 14) == 0)
		return 0;
	/* invalid bitrate */
	if (extract_bitfield(h, 16, 19) == 0xF)
		return 0;
	/* invalid frequency */
	if (extract_bitfield(h, 20, 21) == 0x3)
		return 0;
	return 1;
}

#define SKIP_BUF (1024*64)
static int mp3ai_getinfo(unsigned char *filename, struct mp3_info *i, unsigned int size)
{
	int fd = -1;
	int n_read;
	unsigned char hdr[MPGHDR_LEN];
	unsigned char buf[SKIP_BUF];
	unsigned char tag[ID3_LEN];
	int j;

	memset(tag, 0, ID3_LEN);
	memset(i, 0, sizeof(*i));

	/* Open the file, then try to read the mpeg header and ID3 tag */
	fd = open(filename, O_RDONLY);
	if (fd == -1) {
		NOTICE("Unable to open '%s': %s", filename, ERRMSG());
		goto err;
	}
	/* Get the file length */
	i->filelen = size;
again:
	n_read = read(fd, hdr, 4);
	if (n_read == -1) {
		NOTICE("Unable to read from '%s': %s", filename, ERRMSG());
		goto err;
	}
	if (n_read != 4) { /* truncated file? */
		NOTICE("%s: Truncated mp3file?", filename);
		goto err;
	}
	/* Check for RIFF header in mpeg files... */
	if (memcmp(hdr, "RIFF", 4) == 0) {
		/* try to seek to the 'data' chunk */
		n_read = read(fd, buf, SKIP_BUF);
		if (n_read == -1) {
			NOTICE("Unable to read from '%s': %s", filename, ERRMSG());
			goto err;
		} else if (n_read < 4) {
			goto err;
		}
		n_read -= 4;
		for(j = 0; j <= n_read; j++) {
			if (memcmp(buf+j, "data", 4) == 0) {
				if (lseek(fd, j+12, SEEK_SET) == -1) {
					NOTICE("Unable to seek in file '%s': %s", filename, ERRMSG());
					goto err;
				}
				goto again;
			}
		}
		/* Oops... can't found the 'data' chunk */
		NOTICE("%s: Unable to find the 'data' chunk in this RIFF file", filename);
		goto err;
	}
	if (!seems_mpeg_hdr(hdr)) {
		/* try to find something like a valid header */
		n_read = read(fd, buf, SKIP_BUF);
		if (n_read == -1) {
			NOTICE("Unable to read from '%s': %s", filename, ERRMSG());
			goto err;
		} else if (n_read < 4) {
			goto err;
		}
		n_read -= 4;
		for(j = 0; j <= n_read; j++) {
			if (seems_mpeg_hdr(buf+j)) {
				memcpy(hdr, buf+j, 4);
				goto ok;
			}
		}
		/* Oops... can't found the 'data' chunk */
		NOTICE("%s: Unable to find a valid mpeg header", filename);
		goto err;
	}
ok:
	if (lseek(fd, i->filelen-ID3_LEN, SEEK_SET) == -1) {
		NOTICE("Unable to seek in file '%s': %s", filename, ERRMSG());
		goto err;
	}
	n_read = read(fd, tag, ID3_LEN);
	if (n_read == -1) {
		NOTICE("Unable to read from '%s': %s", filename, ERRMSG());
		goto err;
	}
	close(fd);
	mp3ai_dechdr(hdr, i); /* Decode header info */
	mp3ai_dectag(tag, i); /* Extract fileds from ID3 tag */
	context;
	return 0;
err:
	if (fd != -1)
		close(fd);
	return 1;
}
