#include "autoconf.h"

#ifdef VORBIS
#include "config.h"
#include "macros.h"
#include "share.h"
#include <vorbis/vorbisfile.h>

/* Ahh... vorbisfile rules. */

int ogg_fill_info(struct sharestruct *share, unsigned char *file)
{
	OggVorbis_File vf;
	vorbis_comment *vc;
	vorbis_info *vi;
	FILE *fd;
	char **c;
	int r;

	context;

	fd = fopen(file, "r");

	r = ov_open(fd, &vf, NULL, 0);
	if(r < 0)
	{
		switch(r)
		{
		case OV_EREAD:
			NOTICE("%s: A read from media returned an error.", file);
			break;
		case OV_ENOTVORBIS:
			NOTICE("%s: Bitstream is not Vorbis data.", file);
			break;
		case OV_EVERSION:
			NOTICE("%s: Vorbis version mismatch.", file);
			break;
		case OV_EBADHEADER:
			NOTICE("%s: Invalid Vorbis bitstream header.", file);
			break;
		case OV_EFAULT:
			NOTICE("%s: Internal logic fault; indicates a bug or heap/stack corruption.", file);
			break;
		}
		fclose(fd);
		return 0;
	}

	vi = ov_info(&vf, -1);
	if(!vi)
	{
		NOTICE("%s: Could not read info.", file);
		ov_clear(&vf);
		return 0;
	}

	vc = ov_comment(&vf, -1);
	if(!vc)
	{
		NOTICE("%s: Could not read comments.", file);
		ov_clear(&vf);
		return 0;
	}

	share->bitrate = vi->bitrate_nominal/1000;
	r = ov_time_total(&vf, -1);
	if(r < 0)
		r = 0;
	share->seclength = r;

	r = 0;
	c = vc->user_comments;
	while(*c)
	{
		if(!strncmp(*c, "TITLE=", 6)) {
			share->id3v2_title = ag_strdup(*c+6); r++;
		} else if(!strncmp(*c, "ARTIST=", 7)) {
			share->id3v2_artist = ag_strdup(*c+7); r++;
		} else if(!strncmp(*c, "ALBUM=", 6)) {
			share->id3v2_album = ag_strdup(*c+6); r++;
		} else if(!strncmp(*c, "TRACKNUMBER=", 12)) {
			share->id3v2_track = atoi(*c+12); r++;
		}
		c++;
	}
	if(r)
		share->have_id3v2 = 1;

	ov_clear(&vf);

	return 1;
}

#endif /* VORBIS */
