#include "config.h"
#include "macros.h"
#include "md5.h"

/*#include "../malloc_debug.c"*/


void *_ag_malloc(int size, unsigned char *file, int line)
{
	static void *ptr;

	if(size > 32768)
		return NULL;

	ptr = malloc(size);
	if(!ptr)
		NOTICE("Unable to allocate memory: %s (at %s:%d)", ERRMSG(), file, line);

	return ptr;
}


void _ag_free(void *ptr, unsigned char *file, int line)
{
	if(!ptr)
	{
		NOTICE("Tried to free() NULL ptr! (at %s:%d)", file, line);
		return;
	}
	free(ptr);

	return;
}


unsigned char *_ag_strdup(const unsigned char *s, unsigned char *file, int line)
{
	unsigned char *ptr;

	if(!s)
	{
		NOTICE("Tried to strdup() NULL ptr! (at %s:%d)", file, line);
		return NULL;
	}

	ptr = strdup(s);
	if(!ptr)
	{
		NOTICE("Unable to strdup(): %s (at%s:%d)", ERRMSG(), file, line);
		return NULL;
	}

	return ptr;
}


/*
	libc gayness - I quote:
	"if there is no  null byte among the first n bytes of src, the result will not be null-terminated."
*/
unsigned char *ag_strncpy(unsigned char *dst, const unsigned char *src, int n)
{
	strncpy(dst, src, n);
	if(n) dst[n-1] = 0; else dst[0] = 0;

	return dst;
}


unsigned char *MD5(const unsigned char *str, int len)
{
	md5_state_t pms;
	static unsigned char md5sum[17];

	md5_init(&pms);
	md5_append(&pms, str, len);
	md5_finish(&pms, md5sum);
	md5sum[16] = 0;

	return md5sum;
}


unsigned char *float2str(float v)
{
	static unsigned char str[32];

	snprintf(str, sizeof(str), "%.2f", v);
	return str;
}


unsigned char *int2str(int v)
{
	static unsigned char str[32];

	snprintf(str, sizeof(str), "%d", v);
	return str;
}


/* preappends "~/.fags/" to a string */
unsigned char *etc_file(unsigned char *path)
{
	static unsigned char file[MAX_PATH];
	unsigned char *home;

	home = getenv("HOME");

	if(snprintf(file, sizeof(file), "%s/.fags/%s", home?home:(unsigned char*)".", path) == -1)
	{
		NOTICE("Path too long!");
		return NULL;
	}

	return file;
}


/* checks if a file exists :) */
int file_exists(unsigned char *str)
{
	FILE *fd;

	fd = fopen(str, "r");
	if(!fd && errno == ENOENT)	/* file doesn't exist */
		return 0;

	if(fd)
		fclose(fd);

	return 1;
}


/* divides a with b and rounds the result - written by arachnid, the math professor :) */
int rdiv(int a, int b)
{
	int v = 0;

	if(b == 0)
		return 0;

	v = a / b;

	if(b % 2)
	{
		if((a % b) >= ((b / 2) + 1))
			v++;
	}
	else if((a % b) >= (b / 2))
	{
		v++;
	}

	return v;
}
